#!/usr/local/bin/ruby
#
# DNS Trick
#
# By:  ͵ <yokota@netlab.is.tsukuba.ac.jp>
#
# $Id: dnstrick.rb,v 1.4 2001/09/03 13:41:18 elca Exp $

require 'socket'
require 'thread'
require 'getopts'

require 'datatype.rb'
require 'util.rb'
require 'hostsearch.rb'

require 'multilog.rb'

$ml = MultiLog.new
$ml.open()

getopts("", "downlink:127.0.0.1", "uplink:192.168.0.1")
$ml.log("DNS downlink: " + $OPT_downlink)
$ml.log("DNS uplink:   " + $OPT_uplink)

trap(1) {p $stat}

def parse_rr(packet, pos)
  buf = ""
  real_len = 0
  flag = FALSE

  while TRUE
    if (packet[pos] & 0xc0) != 0
      pos = (packet[pos,2].unpack("n")[0]) & 0x3fff
      if flag == FALSE
	real_len += 2
      end
      flag = TRUE

    elsif packet[pos] != 0
      len = packet[pos]+1
      buf += packet[pos,len]
      pos += len
      if flag == FALSE
	real_len += len
      end

    else
      buf += packet[pos,1]
      if flag == FALSE
	real_len += 1
      end

      break

    end
  end

  ret = {
    "data" => buf,
    "real_len" => real_len
  }

#p ret
  return ret
end


def parse_answer(packet, pos)
  ret = []

  while TRUE
    if pos >= packet.length
      break
    end

    rr = parse_rr(packet, pos)
    len = rr["real_len"]

    a        = rr["data"]

    a_type   = packet[pos+len,    2]
    a_class  = packet[pos+len+2,  2]
    ttl      = packet[pos+len+4,  4]
    data_len = packet[pos+len+8,  2].unpack("n")[0]
    data     = packet[pos+len+10, data_len]

    ret += [{
      "a" => a,
      "a_type" => a_type,
      "a_class" => a_class,
      "ttl" => ttl,
      "data" => data
    }]

    pos += len + 2 + 2 + 4 + 2 + data_len
  end

#p ret
  return ret
end

def parse_packet(packet)
  (id, flags, num_q, ans_rr, ort_rr, add_rr, str) =  packet.unpack("a2 a2 a2 a2 a2 a2 a*")

  if num_q.unpack("n")[0] != 1
    return nil
  end

  rr = parse_rr(packet, 12)
  rr_len = rr["real_len"]

  (q, q_type, q_class, answer) = str.unpack("a#{rr_len.to_s} a2 a2 a*")

  ret = {
    #"packet" => packet,
    "id" => id,
    "flags" => flags,
    "num_q" => num_q,
    "ans_rr" => ans_rr,
    "ort_rr" => ort_rr,
    "add_rr" => add_rr,
    "query" => [
      {
	"q" => q,
	"q_type" => q_type,
	"q_class" => q_class
      }
    ],
  }

#  p rr_len
#  p ret
  if ans_rr.unpack("n")[0] != 0
    ret["answer"] = parse_answer(packet, 12+rr_len+2+2)
  end

  return ret
end

def reject_packet(packet, addr)
  r = packet
  r[2] |= 0x80  # answer
  r[2] &= ~0x04 # not authenticated
  r[3] &= ~0x0f
  r[3] |= 0x05  # query refused error

  $downlink.send(r, 0, addr)
end

def make_key(p)
  return p["id"]+p["query"][0]["q"]
end

#############################################################################

$downlink = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)
$downlink.bind([Socket::AF_INET, 53, str_to_ipstr($OPT_downlink), ""].pack("s n a4 a8"))
$uplink = Socket.open(Socket::AF_INET, Socket::SOCK_DGRAM, 0)

$session_cache = Hash.new
$addr_cache = Hash.new

# stat
$stat = {
  "allreq" => 0,
  "allreply" => 0,
  "numhosts" => Array.new(40,0),
  "cached" => 0,
  "cache_use" => 0,
}

# 饤Ȥ׵äƾ̥Ф
Thread.start do
  loop do
    (client_packet, client_addr) = $downlink.recvfrom(1024)
    $stat["allreq"] += 1

    Thread.start do

      $ml.log("q:  " + client_packet.dump)
      #p client_addr

      p = parse_packet(client_packet)
      if p == nil
	$ml.log("rej:" + client_packet.dump)
	reject_packet(client_packet, client_addr)
	Thread.current.exit
      end


      # å夬¸ߤо̥Ф䤤碌뤳Ȥʤʬֻ֤
      if p["query"][0]["q_type"] == DnsType::A &&
	  $addr_cache[p["query"][0]["q"]] != nil
	#p "use cache"
	$stat["cache_use"] += 1

	r = client_packet[0,12] + 
	  p["query"][0]["q"] + 
	  p["query"][0]["q_type"] + 
	  p["query"][0]["q_class"]
	r[2] |= 0x80  # answer
	r[2] &= ~0x04 # not authenticated
	r[3] &= ~0x0f # no error

	ptr = $addr_cache[p["query"][0]["q"]]
	ttl = max(ptr["timeout"] - Time.now.to_i, 0)

	array_randomize(ptr["addr"]).each { |i|
	  r += "\xc0\x0c" + DnsType::A + DnsClass::INET +
	    [ttl].pack("N") +
	    "\0\4" + i
	}
	r[6,2] = [ptr["addr"].size].pack("n")

	$ml.log("cac:" + r.dump)
	$downlink.send(r, 0, client_addr)

	Thread.current.exit

      end

      $session_cache[make_key(p)] = {
	"addr" => client_addr,
	"timeout" => Time.now.to_i + 60
      }

      $uplink.send(client_packet, 0, [Socket::AF_INET, 53, str_to_ipstr($OPT_uplink), ""].pack("s n a4 a8"))

    end
  end
end


# ̥ФȤֻäƥ饤Ȥ
Thread.start do
  loop do
    (server_packet, server_addr) = $uplink.recvfrom(1024)
    $stat["allreply"] += 1

    Thread.start do

      p = parse_packet(server_packet)
      key = make_key(p)
      if $session_cache[key] == nil
	$ml.log("server reply not found in session cache: " + server_packet.dump)
	Thread.current.exit
      end

      client_addr = $session_cache[key]["addr"]
      q      = p["query"][0]["q"]
      q_type = p["query"][0]["q_type"]

      if (server_packet[2] & 0x02) != 0
	$ml.log("xxx: clear error. FIXME")
	server_packet[2] &= ~0x02 # xxx : must support TCP sever for long query
      end
      $ml.log("ans:" + server_packet.dump)
      $downlink.send(server_packet, 0, client_addr)

      if p["answer"] == nil
	Thread.current.exit
      end

      # p["ans_rr"] μϤɤ⿮ѤǤʤ褦
      num_a = 0
      p["answer"].each {|i|
	if i["a"] == p["query"][0]["q"] &&
	    i["a_type"] == p["query"][0]["q_type"]
	  num_a += 1
	end
      }
      $stat["numhosts"][num_a] += 1

      #p p["answer"]

      if (q_type == DnsType::A || q_type == DnsType::ANY) &&
	  num_a > 1 &&
	  $addr_cache[q] == nil
	$ml.log("create cache: " + dnsstr_to_str(q))

	hosts = []
	p["answer"].each { |i|
	  if i["a_type"]       == DnsType::A         &&
	      i["a_class"]     == DnsClass::INET     &&
	      i["a"]           == p["query"][0]["q"] &&
	      i["data"].length == 4
	    hosts += [i["data"]]
	  end
	}
	if hosts.size == 0
	  Thread.current.exit
	end

	$addr_cache[q] = {
	  "addr" => hosts,
	  "timeout" => Time.now.to_i + min(900, p["answer"][0]["ttl"].unpack("N")[0])
	}

	# ᤤۥȤĴ٤륹å
	Thread.start do
	  $ml.log("start search " + ipstr_array_to_str($addr_cache[q]["addr"]))
	  s = HostSearch.new($addr_cache[q]["addr"])
	  l = s.search(1)
	  s.close
	  if l == nil
	    $ml.log("search fail " + q.dump)
	    # $addr_cache.delete(q)
	  else
	    $stat["cached"] += 1
	    $ml.log("select host: " + dnsstr_to_str(q) + " / " + ipstr_array_to_str(l))
	    $addr_cache[q]["addr"] = l
	  end
	end

      end

      $session_cache.delete(key)

    end
  end
end


# å礭
#
# å "timeout" ǤäƤ
def cache_reduction(cache, overflow_size, logstr)
  h = Hash.new
  cache.each_key {|key|
    h[cache[key]["timeout"]] = key
  }
  if h.size > overflow_size
    del_h = h.keys.sort[0, h.size - overflow_size] # Ť¤٤
    del_h.each { |host|
      $ml.log(logstr + h[host].dump)
      cache.delete(h[host])
    }
  end
end


# Ťǡõ
loop do
  sleep 5

  $session_cache.each_key {|key|
    #p $session_cache
    if $session_cache[key]["timeout"] < Time.now.to_i
      $ml.log("query timeout: key=" + key.dump)
      $session_cache.delete(key)
    end
  }

  $addr_cache.each_key {|key|
    #p $addr_cache
    if $addr_cache[key]["timeout"] < Time.now.to_i
      $ml.log("cache timeout: " + dnsstr_to_str(key))
      $addr_cache.delete(key)
    end
  }

  # å˥¤򤫤
  cache_reduction($session_cache, 500, "query cache overflow: key=")
  cache_reduction($addr_cache, 500, "address cache overflow: key=")

end

# end
